@@ -6,14 +6,15 @@ dnfile, .NET Executable file parser
 REFERENCES
 
     https://www.ntcore.com/files/dotnetformat.htm
-    https://referencesource.microsoft.com/System.AddIn/System/Addin/MiniReflection/MetadataReader/Metadata.cs.html#123
+    https://referencesource.microsoft.com/System.AddIn/System/Addin/MiniReflection/MetadataReader/Metadata.cs.html
+    https://asmresolver.readthedocs.io/en/latest/peimage/dotnet.html#metadata-streams
 
 
 Copyright (c) 2020-2022 MalwareFrank
 """
 
 __author__ = """MalwareFrank"""
-__version__ = "0.10.0"
+__version__ = "0.15.1"
 
 import copy as _copymod
 import codecs
@@ -24,7 +25,7 @@ from typing import Dict, List, Optional
 from pefile import PE as _PE
 from pefile import DIRECTORY_ENTRY, MAX_SYMBOL_EXPORT_COUNT, Dump, Structure, DataContainer, PEFormatError
 
-from . import base, enums, errors, stream
+from . import base, enums, errors, stream, resource
 
 logger = logging.getLogger(__name__)
 CLR_METADATA_SIGNATURE = 0x424A5342
@@ -56,8 +57,10 @@ class dnPE(_PE):
         data=None,
         fast_load=None,
         max_symbol_exports=MAX_SYMBOL_EXPORT_COUNT,
+        clr_lazy_load=False,
     ):
         self._warnings = list()
+        self.clr_lazy_load = clr_lazy_load
         super().__init__(name, data, fast_load)
 
     def dump_info(self, dump=None, encoding="utf-8"):
@@ -215,7 +218,7 @@ class dnPE(_PE):
 
     def parse_clr_structure(self, rva, size) -> Optional["ClrData"]:
         try:
-            return ClrData(self, rva, size)
+            return ClrData(self, rva, size, self.clr_lazy_load)
         except errors.dnFormatError as e:
             logger.warning("failed to parse CLR data: %s",  e)
             return None
@@ -270,7 +273,7 @@ class ClrMetaData(DataContainer):
     # dw    NumberOfStreams
     # var   StreamHeaders
 
-    def __init__(self, pe: dnPE, rva: int, size: int):
+    def __init__(self, pe: dnPE, rva: int, size: int, lazy_load=False):
         """
         Given a dnPE object, MetaData RVA and MetaData Size.
         Raises dnFormatError if encounter problems parsing.
@@ -339,7 +342,7 @@ class ClrMetaData(DataContainer):
             # parse each stream
             for s in self.streams_list:
                 try:
-                    s.parse(self.streams_list)
+                    s.parse(self.streams_list, lazy_load=lazy_load)
                 except (errors.dnFormatError, PEFormatError) as e:
                     # other streams may parse, so add to warnings and continue
                     pe.add_warning("Unable to parse stream {!r}".format(s.struct.Name))
@@ -422,6 +425,15 @@ class ClrData(DataContainer):
     mdtables: Optional[stream.MetaDataTables]
     Flags: Optional[enums.ClrHeaderFlags]
 
+    _resources: Optional[List[base.ClrResource]]
+
+    @property
+    def resources(self) -> List[base.ClrResource]:
+        if self._resources is None:
+            self._init_resources(getattr(self, "_pe"))
+            assert self._resources is not None
+        return self._resources
+
     # Structure description from:
     # http://www.ntcore.com/files/dotnetformat.htm
     _format = (
@@ -449,7 +461,7 @@ class ClrData(DataContainer):
         ),
     )
 
-    def __init__(self, pe: dnPE, rva: int, size: int):
+    def __init__(self, pe: dnPE, rva: int, size: int, lazy_load: bool):
         """
         Given dnPE object, .NET header RVA and header size.
         Raises dnFormatError if problems parsing.
@@ -481,10 +493,11 @@ class ClrData(DataContainer):
         self.guids = None
         self.blobs = None
         self.mdtables = None
+        self._resources = None
         self.Flags = None
 
         try:
-            self.metadata = ClrMetaData(pe, metadata_rva, metadata_size)
+            self.metadata = ClrMetaData(pe, metadata_rva, metadata_size, lazy_load)
         except (errors.dnFormatError, PEFormatError) as e:
             logger.warning("failed to parse .NET metadata: %s", e)
             return
@@ -509,11 +522,68 @@ class ClrData(DataContainer):
         flags_object = enums.ClrHeaderFlags(clr_struct.Flags)
         self.Flags = flags_object
 
+        if not lazy_load:
+            self._init_resources(pe)
+        else:
+            # store the dnPE reference for lazy-loading
+            setattr(self, "_pe", pe)
+
+    def _init_resources(self, pe):
+        """Parse and initialize assembly resources.
+
+        This is separate from `ClrData.__init__` to allow for mdtable lazy-loading since
+        parsing `ManifestResourceRow.Implementation` requires all tables to be loaded.
+        """
+        self._resources = []
+        # parse the resources
+        if self.struct.ResourcesRva > 0 and self.mdtables and self.mdtables.ManifestResource and self.mdtables.ManifestResource.num_rows > 0:
+            # for each row
+            for row in self.mdtables.ManifestResource.rows:
+                # TODO: handle external resources
+                if row.Implementation is None:
+                    # internal resource, embedded in this file
+                    rva = self.struct.ResourcesRva + row.Offset
+                    try:
+                        buf = pe.get_data(rva, 4)
+                    except PEFormatError as e:
+                        # warn
+                        pe.add_warning("CLR resource parse error, expected more data at rva 0x{:02x}".format(rva))
+                        continue
+                    if not buf or len(buf) < 4:
+                        # warn
+                        pe.add_warning("CLR resource parse error, expected at least 4 bytes at rva 0x{:02x}".format(rva))
+                        continue
+                    size = int.from_bytes(buf, byteorder="little")
+                    rsrc_rva = rva + 4
+                    try:
+                        rdata = pe.get_data(rsrc_rva, size)
+                    except PEFormatError as e:
+                        # warn
+                        pe.add_warning("CLR resource parse error, expected more data at rva 0x{:02x}".format(rsrc_rva))
+                        continue
+                    if not rdata or len(rdata) < size:
+                        pe.add_warning("CLR resource parse error, expected more data at rva 0x{:02x}".format(rsrc_rva))
+                        continue
+                    res = resource.InternalResource(row.Name, row.Flags.mrPublic, row.Flags.mrPrivate)
+                    res.rva = rsrc_rva
+                    res.size = size
+                    res.data = rdata
+                    self._resources.append(res)
+            for rsrc in self._resources:
+                try:
+                    rsrc.parse()
+                except errors.dnFormatError as e:
+                    if isinstance(rsrc, resource.InternalResource):
+                        pe.add_warning("CLR resource parse error for '{}' at 0x{:02x}: {}".format(rsrc.name, rsrc.rva, str(e)))
+                    else:
+                        pe.add_warning("CLR resource parse error for '{}': {}".format(rsrc.name, str(e)))
+
 
 class ClrStreamFactory(object):
     _name_type_map = {
         b"#~": stream.MetaDataTables,
         b"#-": stream.MetaDataTables,
+        b"#Schema": stream.MetaDataTables,
         b"#Strings": stream.StringsHeap,
         b"#GUID": stream.GuidHeap,
         b"#Blob": stream.BlobHeap,
@@ -554,8 +624,9 @@ class ClrStreamFactory(object):
         stream_struct.__unpack__(struct_data)
         # remove trailing NULLs from name
         stream_struct.Name = stream_struct.Name.rstrip(b"\x00")
+        stream_rva = metadata_rva + stream_struct.Offset
         stream_data = pe.get_data(
-            metadata_rva + stream_struct.Offset, stream_struct.Size
+            stream_rva, stream_struct.Size
         )
         name = stream_struct.Name
         # use GenericStream for any non-standard streams
@@ -563,6 +634,7 @@ class ClrStreamFactory(object):
         try:
             # construct stream, like stream.StreagsHeap ctor or GenericStream ctor
             s = stream_class(metadata_rva, stream_struct, stream_data)
+            s.file_offset = pe.get_offset_from_rva(stream_rva)
         except errors.dnFormatError as e:
             logger.warning("failed to parse stream: %s", e)
             return None
