@@ -2,18 +2,21 @@
 """
 .NET base classes
 
-Copyright (c) 2020-2022 MalwareFrank
+Copyright (c) 2020-2024 MalwareFrank
 """
 import abc
 import enum
 import struct as _struct
-import typing
 import logging
-from typing import TYPE_CHECKING, Dict, List, Type, Tuple, Union, Generic, TypeVar, Optional, Sequence
+import functools as _functools
+import itertools as _itertools
+from typing import TYPE_CHECKING, Any, Dict, List, Type, Tuple, Union, Generic, TypeVar, Optional, Sequence
 
 from pefile import Structure
 
 from . import enums, errors
+from .utils import LazyList as _LazyList
+from .utils import read_compressed_int as _read_compressed_int
 
 if TYPE_CHECKING:
     from . import stream
@@ -22,6 +25,28 @@ if TYPE_CHECKING:
 logger = logging.getLogger(__name__)
 
 
+class CompressedInt(int):
+    raw_size: int
+    __data__: bytes
+    value: int
+    rva: Optional[int] = None
+
+    def to_bytes(self):
+        return self.__data__
+
+    @classmethod
+    def read(cls, data: bytes, rva: Optional[int] = None) -> Optional["CompressedInt"]:
+        result = _read_compressed_int(data)
+        if result is None:
+            return None
+        ci = CompressedInt(result[0])
+        ci.raw_size = result[1]
+        ci.value = result[0]
+        ci.__data__ = data[:result[1]]
+        ci.rva = rva
+        return ci
+
+
 class StreamStruct(Structure):
     Name: bytes
     Offset: int
@@ -37,11 +62,12 @@ class ClrStream(abc.ABC):
     ):
         self.struct: StreamStruct = stream_struct
         self.rva: int = metadata_rva + stream_struct.Offset
+        self.file_offset: Optional[int] = None
         self.__data__: bytes = stream_data
         self._stream_table_entry_size = stream_struct.sizeof()
         self._data_size = len(stream_data)
 
-    def parse(self, streams: List):
+    def parse(self, streams: List, lazy_load: bool = False):
         """
         Parse the stream.
 
@@ -62,6 +88,17 @@ class ClrStream(abc.ABC):
         """
         return self._data_size
 
+    def get_file_offset(self, rva=None):
+        """
+        Return the file offset of the given RVA within this stream.
+        If no RVA given, then return the file offset of this stream.
+        """
+        if self.file_offset is not None and rva is not None:
+            stream_offset = rva - self.rva
+            if stream_offset >= 0:
+                return self.file_offset + stream_offset
+        return self.file_offset
+
     def get_data_at_offset(self, offset, size):
         if size == 0 or offset >= self.sizeof():
             return b""
@@ -80,9 +117,71 @@ class ClrStream(abc.ABC):
         return i
 
 
+class HeapItem(abc.ABC):
+    """
+    HeapItem is a base class for items retrieved from any of the
+    heap streams, for example #Strings, #US, #GUID, and #Blob.
+
+    It can be used to access the raw underlying data, the RVA
+    from which it was retrieved, an optional interpreted value,
+    and the bytes representation of the value.
+
+    Each heap stream .get() call returns a subclass with these
+    and optionally additional members.
+    """
+
+    rva: Optional[int] = None
+    # original data from file
+    __data__: bytes
+    # interpreted value
+    value: Any = None
+
+    def __init__(self, data: bytes, rva: Optional[int] = None):
+        self.rva = rva
+        self.__data__ = data
+
+    def value_bytes(self):
+        """
+        Return the raw bytes underlying the interpreted value.
+
+        For the base HeapItem, this is the same as the raw_data.
+        """
+        return self.__data__
+
+    @property
+    def raw_size(self):
+        """
+        Number of bytes read from the stream, including any header,
+        value, and footer.
+        """
+        return len(self.__data__)
+
+    @property
+    def raw_data(self):
+        """
+        The bytes read from the stream, including any header,
+        value, and footer
+        """
+        return self.__data__
+
+    def __eq__(self, other):
+        """
+        Two HeapItems are equal if their raw data is the same or their
+        interpreted values are the same and not Noney.
+
+        A HeapItem is equal to a bytes object if the HeapItem's value as bytes
+        is equal to the bytes object.
+        """
+        if isinstance(other, HeapItem):
+            return self.raw_data == other.raw_data or (self.value is not None and self.value == other.value)
+        elif isinstance(other, bytes):
+            return self.value_bytes() == other
+        return False
+
+
 class ClrHeap(ClrStream):
     @abc.abstractmethod
-    def get(self, index):
+    def get(self, index: int):
         raise NotImplementedError()
 
 
@@ -90,6 +189,12 @@ class RowStruct(Structure):
     pass
 
 
+class LoadState(enum.Enum):
+    Unloaded = 0
+    LazyLoaded = 1
+    Loaded = 2
+
+
 class MDTableRow(abc.ABC):
     """
     This is the base class for Metadata Tables' rows.
@@ -176,6 +281,8 @@ class MDTableRow(abc.ABC):
         """
         assert hasattr(self.__class__, "_struct_class")
 
+        self._loaded = LoadState.Unloaded
+
         self._tables_rowcnt = tables_rowcounts
         self._strings: Optional["stream.StringsHeap"] = strings_heap
         self._guids: Optional["stream.GuidHeap"] = guid_heap
@@ -201,7 +308,7 @@ class MDTableRow(abc.ABC):
         """
         ...
 
-    def set_data(self, data: bytes, offset: int = None):
+    def set_data(self, data: bytes, file_offset: Optional[int] = None):
         """
         Parse the data and set struct for this row.
 
@@ -209,9 +316,60 @@ class MDTableRow(abc.ABC):
         parse() is called after all tables have had parse_rows() called on them.
         """
         self._data = data
-        self.struct = self.__class__._struct_class(format=self._format, file_offset=offset)
+        self.struct = self.__class__._struct_class(format=self._format, file_offset=file_offset)
         self.struct.__unpack__(data)
 
+    # can be safely parsed without all tables being initialized
+    CLASS_ATTRS = (
+        "_struct_asis", "_struct_strings", "_struct_guids",
+        "_struct_blobs", "_struct_flags", "_struct_enums",
+
+    )
+    # cannot be fully parsed without all tables being initialized
+    CLASS_ATTRS_TABLES = (
+        "_struct_codedindexes", "_struct_indexes", "_struct_lists"
+    )
+
+    def setup_lazy_load(self, full_loader):
+        """Mark this row for lazy-loading.
+
+        `full_loader` will be called if a property is requested that requires
+        loading all mdtables before it can be parsed.
+        """
+        if not self._loaded == LoadState.Unloaded:
+            return
+        self._loaded = LoadState.LazyLoaded
+        self._full_loader = full_loader
+        # Retrieve the properties that this row *could* have, along with
+        # their associated struct type. These are used to determine what
+        # _parse function needs to be called to load the property.
+        self._class_struct_attrs, self._class_struct_attrs_tables = (
+            _row_class_struct_attrs(self.__class__)
+        )
+
+    def __getattr__(self, attr):
+        """If this row is marked for lazy-loading, attempt to load the struct
+        that contains the requested property and try again.
+
+        If the requested property requires data from other
+        mdtables, all tables will be loaded.
+        """
+        if self._loaded == LoadState.LazyLoaded:
+            if attr in self._class_struct_attrs:
+                loader = getattr(self, "_parse" + self._class_struct_attrs[attr])
+                loader()
+                # If something were to go wrong with loading the correct struct, this
+                # would cause a StackOverflow from recursive __getattr__ calls.
+                if hasattr(self, attr):
+                    return getattr(self, attr)
+            elif attr in self._class_struct_attrs_tables:
+                # This property requires data from other tables, trigger a full load.
+                self._full_loader()
+                if hasattr(self, attr):
+                    return getattr(self, attr)
+        # TODO: more descriptive exception?
+        raise AttributeError(attr)
+
     def parse(self, tables: List["ClrMetaDataTable"], next_row: Optional["MDTableRow"]):
         """
         Parse the row data and set object attributes.  Should only be called after all rows of all tables
@@ -219,12 +377,25 @@ class MDTableRow(abc.ABC):
 
             next_row    the next row in the table, used for row lists (e.g. FieldList, MethodList)
         """
+        self._parse_struct_asis()
+        self._parse_struct_strings()
+        self._parse_struct_guids()
+        self._parse_struct_blobs()
+        self._parse_struct_flags()
+        self._parse_struct_enums()
+        self._parse_struct_codedindexes(tables, next_row)
+        self._parse_struct_indexes(tables, next_row)
+        self._parse_struct_lists(tables, next_row)
+        self._loaded = LoadState.Loaded
+
+    def _parse_struct_asis(self):
         # if there are any fields to copy as-is
         if hasattr(self.__class__, "_struct_asis"):
             for struct_name, attr_name in self.__class__._struct_asis.items():
                 # always define attribute, even if failed to parse
                 setattr(self, attr_name, getattr(self.struct, struct_name, None))
 
+    def _parse_struct_strings(self):
         # if strings
         if hasattr(self.__class__, "_struct_strings"):
             for struct_name, attr_name in self.__class__._struct_strings.items():
@@ -245,6 +416,7 @@ class MDTableRow(abc.ABC):
                 except IndexError:
                     logger.warning("failed to fetch string: unable to parse data")
 
+    def _parse_struct_guids(self):
         # if guids
         if hasattr(self.__class__, "_struct_guids"):
             for struct_name, attr_name in self.__class__._struct_guids.items():
@@ -260,6 +432,7 @@ class MDTableRow(abc.ABC):
                 except (IndexError, TypeError):
                     logger.warning("failed to fetch guid: unable to parse data")
 
+    def _parse_struct_blobs(self):
         # if blobs
         if hasattr(self.__class__, "_struct_blobs"):
             for struct_name, attr_name in self.__class__._struct_blobs.items():
@@ -274,6 +447,7 @@ class MDTableRow(abc.ABC):
                 except (IndexError, TypeError):
                     logger.warning("failed to fetch blob: unable to parse data")
 
+    def _parse_struct_codedindexes(self, tables, next_row):
         # if coded indexes
         if hasattr(self.__class__, "_struct_codedindexes") and tables:
             for struct_name, (
@@ -288,6 +462,7 @@ class MDTableRow(abc.ABC):
                 except (IndexError, TypeError):
                     logger.warning("failed to fetch coded index: unable to parse data")
 
+    def _parse_struct_flags(self):
         # if flags
         if hasattr(self.__class__, "_struct_flags"):
             for struct_name, (attr_name, flag_class) in self.__class__._struct_flags.items():
@@ -304,6 +479,7 @@ class MDTableRow(abc.ABC):
                 except ValueError:
                     logger.warning("failed to fetch flag: invalid flag data")
 
+    def _parse_struct_enums(self):
         # if enums
         if hasattr(self.__class__, "_struct_enums"):
             for struct_name, (attr_name, enum_class) in self.__class__._struct_enums.items():
@@ -320,6 +496,7 @@ class MDTableRow(abc.ABC):
                 except ValueError:
                     logger.warning("failed to fetch enum: invalid enum data")
 
+    def _parse_struct_indexes(self, tables, next_row):
         # if indexes
         if hasattr(self.__class__, "_struct_indexes") and tables:
             for struct_name, (attr_name, table_name) in self.__class__._struct_indexes.items():
@@ -337,6 +514,7 @@ class MDTableRow(abc.ABC):
                     else:
                         logger.warning("failed to fetch index reference: unable to parse data")
 
+    def _parse_struct_lists(self, tables, next_row):
         # if lists
         if hasattr(self.__class__, "_struct_lists") and tables:
             for struct_name, (attr_name, table_name) in self.__class__._struct_lists.items():
@@ -367,7 +545,8 @@ class MDTableRow(abc.ABC):
                         next_row_reference = getattr(next_row.struct, struct_name, None)
                         run_end_index = max_row
                         if next_row_reference is not None:
-                            run_end_index = min(next_row_reference, max_row)
+                            # row end index is inclusive so row end index must equal next row index minus 1, if less than max row
+                            run_end_index = min(next_row_reference - 1, max_row)
 
                     else:
                         # then we read from the target table,
@@ -378,8 +557,8 @@ class MDTableRow(abc.ABC):
                     # when this run starts at the last index,
                     # start == end and end == max_row.
                     # otherwise, if start == end, then run is empty.
-                    if (run_start_index != run_end_index) or (run_end_index == max_row):
-                        # row indexes are 1-indexed, so our range goes to end+1
+                    if run_start_index <= run_end_index:
+                        # row indexes are inclusive, so our range goes to end+1
                         for row_index in range(run_start_index, run_end_index + 1):
                             run.append(MDTableIndex(table, row_index))
 
@@ -423,6 +602,29 @@ class MDTableRow(abc.ABC):
             return "I"
 
 
+# Computing this for each class takes some time, especially if it is done for every row,
+# but it *should* remain consistent for any given class and therefore can be cached.
+@_functools.lru_cache(None)
+def _row_class_struct_attrs(cls: Type[MDTableRow]):
+    """Retrieve all possible attributes for a `MDTableRow` class,
+    along with their associated struct type.
+
+    Attributes are separated based on whether they can be loaded without data from
+    any other tables.
+    """
+    attrs = {
+        attr[0] if isinstance(attr, tuple) else attr: struct
+        for struct in MDTableRow.CLASS_ATTRS
+        for _, attr in getattr(cls, struct, {}).items()
+    }
+    attrs_tables = {
+        attr[0] if isinstance(attr, tuple) else attr: struct
+        for struct in MDTableRow.CLASS_ATTRS_TABLES
+        for _, attr in getattr(cls, struct, {}).items()
+    }
+    return attrs, attrs_tables
+
+
 class MDTablesStruct(Structure):
     Reserved_1: int
     MajorVersion: int
@@ -503,7 +705,7 @@ class CodedIndex(MDTableIndex[RowType]):
             self.table = t
             return
 
-        logger.warning("reference to missing table: %s" % table_name)
+        # this may not be a problem, e.g. when ManifestResource Implementation=0
         self.table = None
 
 
@@ -534,6 +736,7 @@ class ClrMetaDataTable(Generic[RowType]):
     #       _row_class = ModuleRow
     #
     rva: int
+    file_offset: int
     number: int
     name: str
     _row_class: Type[RowType]
@@ -548,6 +751,7 @@ class ClrMetaDataTable(Generic[RowType]):
         strings_heap: Optional["stream.StringsHeap"],
         guid_heap: Optional["stream.GuidHeap"],
         blob_heap: Optional["stream.BlobHeap"],
+        lazy_load=False
     ):
         """
         Given the tables' row counts, sorted flag, and heap info.
@@ -562,35 +766,50 @@ class ClrMetaDataTable(Generic[RowType]):
         assert hasattr(self, "name")
         assert hasattr(self, "_row_class")
 
+        self._loaded = LoadState.Unloaded
+
         # default value
         self.rva: int = 0
+        self.file_offset = 0
 
         num_rows = tables_rowcounts[self.number]
         if num_rows is None:
             # the table doesn't exist, so create the instance, but with zero rows.
-            logger.warning("reference to missing table: %d" % (self.number))
             num_rows = 0
 
         self.is_sorted: bool = is_sorted
         self.num_rows: int = num_rows
 
-        self.rows: List[RowType] = []
-        for i in range(num_rows):
+        def init_row():
+            return self._row_class(
+                tables_rowcounts,
+                strings_offset_size,
+                guid_offset_size,
+                blob_offset_size,
+                strings_heap,
+                guid_heap,
+                blob_heap,
+            )
+
+        self.rows: List[RowType]
+        if lazy_load and num_rows > 0:
+            self.rows = _LazyList(self._lazy_parse_rows, num_rows)
             try:
-                self.rows.append(self._row_class(
-                    tables_rowcounts,
-                    strings_offset_size,
-                    guid_offset_size,
-                    blob_offset_size,
-                    strings_heap,
-                    guid_heap,
-                    blob_heap,
-                ))
+                # `_get_row_size` uses the size of the first row.
+                self.rows[0] = init_row()
             except errors.dnFormatError:
-                # this may occur when the offset to a stream is too large.
-                # this probably means invalid data.
-                logger.warning("failed to construct %s row %d", self.name, i)
-                break
+                logger.warning("failed to construct %s row %d", self.name, 0)
+                # "truncate" the list since the following data is assumed invalid.
+                self.rows = []
+        else:
+            self.rows = []
+            for e in range(num_rows):
+                try:
+                    self.rows.append(init_row())
+                except errors.dnFormatError:
+                    # this may occur when the offset to a stream is too large.
+                    # this probably means invalid data.
+                    logger.warning("failed to construct %s row %d", self.name, e)
 
         # store heap info
         self._strings_heap: Optional["stream.StringsHeap"] = strings_heap
@@ -610,6 +829,74 @@ class ClrMetaDataTable(Generic[RowType]):
         r = self.rows[0]
         return r.row_size
 
+    def setup_lazy_load(self, table_rva: int, data: bytes, full_loader):
+        """Mark this table for lazy-loading.
+
+        `full_loader` will be called if a row property is requested that requires
+        loading all mdtables before it can be parsed.
+        """
+        if not self._loaded == LoadState.Unloaded:
+            return
+        self.rva = table_rva
+        self._table_data = data
+        self._full_loader = full_loader
+        if len(data) < self.row_size * self.num_rows:
+            logger.warning("not enough data to parse %d rows", self.num_rows)
+        self._loaded = LoadState.LazyLoaded
+
+    def _lazy_parse_row(self, row, idx):
+        """If the row has not been loaded, return a new lazy-loaded row initialized
+        with the proper data for the row index.
+        """
+        if row and row._loaded != LoadState.Unloaded:
+            return row
+
+        try:
+            row = self._row_class(
+                self._tables_rowcounts,
+                self._strings_offset_size,
+                self._guid_offset_size,
+                self._blob_offset_size,
+                self._strings_heap,
+                self._guid_heap,
+                self._blob_heap,
+            )
+        except errors.dnFormatError:
+            # this may occur when the offset to a stream is too large.
+            # this probably means invalid data.
+            logger.warning("failed to construct %s row %d", self.name, idx)
+            # truncate the row list to the current index since the following data is
+            # assumed invalid.
+            assert isinstance(self.rows, _LazyList)
+            self.rows.truncate(idx)
+            return None
+
+        row.setup_lazy_load(self._full_loader)
+        offset = self.row_size * idx
+        if len(self._table_data) < offset + self.row_size:
+            logger.warning("not enough data to parse row %d", idx)
+            # we could truncate here as well, but regular loading would still be
+            # left with a full-length list in the equivalent situation.
+            return row
+        row.set_data(self._table_data[offset:offset + self.row_size], file_offset=self.file_offset + offset)
+        return row
+
+    def _lazy_parse_rows(self, key, row):
+        """Convenience function to handle both indexing and slicing of the LazyList."""
+        # Guard if called before setup_lazy_load()
+        if self._loaded != LoadState.LazyLoaded:
+            return row
+
+        if isinstance(key, slice):
+            # normally not knowing the length of the list a slice was taken from
+            # would be an issue, but we can just keep going until `row` is exhausted.
+            return [
+                self._lazy_parse_row(row, i)
+                for row, i in zip(row, _itertools.count(key.start or 0, key.step or 1))
+            ]
+
+        return self._lazy_parse_row(row, key)
+
     def parse_rows(self, table_rva: int, data: bytes):
         """
         Given a byte sequence containing the rows, add data to each row in the
@@ -617,6 +904,10 @@ class ClrMetaDataTable(Generic[RowType]):
         parse() is called, which should not happen until all tables have been
         initialized and parse_rows() called on each.
         """
+        if self._loaded == LoadState.LazyLoaded:
+            # will be handled by lazy evaluation during parse()
+            return
+
         self._table_data = data
         if len(data) < self.row_size * self.num_rows:
             logger.warning("not enough data to parse %d rows", self.num_rows)
@@ -630,7 +921,7 @@ class ClrMetaDataTable(Generic[RowType]):
                 break
 
             self.rows[i].set_data(
-                data[offset:offset + self.row_size], offset=table_rva + offset
+                data[offset:offset + self.row_size], file_offset=self.file_offset + offset
             )
             offset += self.row_size
 
@@ -651,6 +942,7 @@ class ClrMetaDataTable(Generic[RowType]):
 
             # fully parse the row
             row.parse(tables, next_row=next_row)
+        self._loaded = LoadState.Loaded
 
     def __getitem__(self, index: int) -> RowType:
         return self.rows[index]
@@ -673,3 +965,18 @@ class ClrMetaDataTable(Generic[RowType]):
         use `__getitem__` when you want 0-based indexing.
         """
         return self[row_index - 1]
+
+
+class ClrResource(abc.ABC):
+    def __init__(self, name: str, public: bool = False, private: bool = False):
+        self.name: str = name
+        self.public: bool = public
+        self.private: bool = private
+        self.data: Optional[bytes] = None
+
+    def set_data(self, data: bytes):
+        self.data = data
+
+    @abc.abstractmethod
+    def parse(self):
+        raise NotImplementedError()
